import logging
from argparse import ArgumentParser

from Bio.Align import PairwiseAligner

from palamedes.align import generate_seq_record, generate_alignment, generate_variant_blocks
from palamedes.config import MOLECULE_TYPE_PROTEIN, ALT_SEQUENCE_ID, REF_SEQUENCE_ID
from palamedes import __version__
from palamedes.hgvs.utils import categorize_variant_block
from palamedes.hgvs.builders import BUILDER_CONFIG
from palamedes.utils import configure_logging
from palamedes.config import (
    GLOBAL_ALIGN_MODE,
    DEFAULT_MATCH_SCORE,
    DEFAULT_MISMATCH_SCORE,
    DEFAULT_OPEN_GAP_SCORE,
    DEFAULT_EXTEND_GAP_SCORE,
)

LOGGER = logging.getLogger(__name__)


def main() -> None:
    parser = ArgumentParser(
        description="Generate HGVS objects for all variants found in the alignment between 2 sequences",
    )
    parser.add_argument(
        "ref",
        help="Reference sequence",
        type=str,
    )

    parser.add_argument(
        "alt",
        help="Alternate sequence",
        type=str,
    )

    parser.add_argument(
        "--ref-id",
        help="Identifier for reference sequence id",
        type=str,
        default=REF_SEQUENCE_ID,
    )
    parser.add_argument(
        "--alt-id",
        help="Identifier for alternate sequence id",
        type=str,
        default=ALT_SEQUENCE_ID,
    )
    parser.add_argument(
        "--molecule-type",
        help="Molecule type to use",
        choices=(MOLECULE_TYPE_PROTEIN,),
        default=MOLECULE_TYPE_PROTEIN,
    )
    parser.add_argument(
        "--use-non-standard-substitution-rules",
        help=(
            "Flag to enable non-standard substitution rules, "
            "replacing balanced delins calls with multiple substitutions"
        ),
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--match-score",
        help="Match score to use for base alignment",
        type=int,
        default=DEFAULT_MATCH_SCORE,
    )
    parser.add_argument(
        "--mismatch-score",
        help="Mismatch score to use for base alignment",
        type=int,
        default=DEFAULT_MISMATCH_SCORE,
    )
    parser.add_argument(
        "--gap-open-score",
        help="Gap open score to use for base alignment",
        type=int,
        default=DEFAULT_OPEN_GAP_SCORE,
    )
    parser.add_argument(
        "--gap-extend-score",
        help="Gap extend score to use for base alignment",
        type=int,
        default=DEFAULT_EXTEND_GAP_SCORE,
    )
    parser.add_argument(
        "--debug",
        help="Enable debug logging to stderr",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    args = parser.parse_args()
    configure_logging(args.debug)

    LOGGER.debug("Running with args: %s", args)

    aligner = PairwiseAligner(
        mode=GLOBAL_ALIGN_MODE,
        match_score=args.match_score,
        mismatch_score=args.mismatch_score,
        open_gap_score=args.gap_open_score,
        extend_gap_score=args.gap_extend_score,
    )

    ref_seq_record = generate_seq_record(args.ref, args.ref_id, molecule_type=args.molecule_type)
    alt_seq_record = generate_seq_record(args.alt, args.alt_id, molecule_type=args.molecule_type)
    alignment = generate_alignment(
        ref_seq_record,
        alt_seq_record,
        molecule_type=args.molecule_type,
        aligner=aligner,
    )

    LOGGER.debug("Found best alignment with score = %s", getattr(alignment, "score"))
    LOGGER.debug("Alignment:\n%s", str(alignment))

    variant_blocks = generate_variant_blocks(
        alignment,
        split_consecutive_mismatches=args.use_non_standard_substitution_rules,
    )
    LOGGER.debug("%s Variant blocks generated", len(variant_blocks))

    builder = BUILDER_CONFIG[args.molecule_type](alignment)
    for variant_block in variant_blocks:
        category = categorize_variant_block(
            variant_block,
            alignment,
        )
        LOGGER.debug("%s, categorized as: %s", variant_block, category)

        hgvs = builder.build(variant_block, category)
        print(hgvs.format())


if __name__ == "__main__":
    main()



from Bio.Align import Alignment

from palamedes.align import get_upstream_reference_sequence
from palamedes.models import VariantBlock
from palamedes.config import (
    HGVS_VARIANT_TYPE_SUBSTITUTION,
    HGVS_VARIANT_TYPE_DELETION,
    HGVS_VARIANT_TYPE_EXTENSION,
    HGVS_VARIANT_TYPE_DUPLICATION,
    HGVS_VARIANT_TYPE_REPEAT,
    HGVS_VARIANT_TYPE_INSERTION,
    HGVS_VARIANT_TYPE_DELETION_INSERTION,
    VARIANT_BASE_MATCH,
    VARIANT_BASE_MISMATCH,
    VARIANT_BASE_DELETION,
    VARIANT_BASE_INSERTION,
)
from palamedes.utils import contains_repeated_substring, yield_repeating_substrings


def categorize_variant_block(variant_block: VariantBlock, alignment: Alignment) -> str:
    """
    Process a variant block (with the global alignment) to categorize it with the correct base HGVS "type".

    The following rule-set is used:
    - If any matches appear in the block => an error is raised
    - A single position mismatch => Substitution
    - 1 or more deletion positions and nothing else => Deletion
    - 1 or more insertion positions and nothing else
        - If fully upstream or fully downstream of reference => Extension
        - If inserted bases perfectly match upstream bases => Duplication
        - If inserted bases are comprised of a repeating sub-string that perfectly matches upstream bases => Repeat
        - If nothing else => Insertion
    - Any other combo of calls is a Deletion-Insertion
    """
    if VARIANT_BASE_MATCH in variant_block.alignment_block.bases:
        raise ValueError(
            f"Cannot categorize VariantBlock with matching bases in it: {variant_block.alignment_block.bases}"
        )

    if variant_block.alignment_block.bases == VARIANT_BASE_MISMATCH:
        return HGVS_VARIANT_TYPE_SUBSTITUTION

    if set(variant_block.alignment_block.bases) == set([VARIANT_BASE_DELETION]):
        return HGVS_VARIANT_TYPE_DELETION

    if set(variant_block.alignment_block.bases) == set([VARIANT_BASE_INSERTION]):
        alignment_end = len(alignment[0])
        if variant_block.alignment_block.start == 0 or variant_block.alignment_block.end == alignment_end:
            return HGVS_VARIANT_TYPE_EXTENSION

        inserted_bases = variant_block.alternate_blocks[0].bases

        if inserted_bases == get_upstream_reference_sequence(
            alignment, variant_block.alignment_block.start, len(inserted_bases)
        ):
            return HGVS_VARIANT_TYPE_DUPLICATION

        if contains_repeated_substring(inserted_bases):
            candidate_repeats = [
                substring
                for substring in yield_repeating_substrings(inserted_bases)
                if get_upstream_reference_sequence(alignment, variant_block.alignment_block.start, len(substring))
                == substring
            ]
            if len(candidate_repeats) > 0:
                return HGVS_VARIANT_TYPE_REPEAT

        return HGVS_VARIANT_TYPE_INSERTION

    return HGVS_VARIANT_TYPE_DELETION_INSERTION



GLOBAL_ALIGN_MODE: str = "global"

# default alignment params, meant to favor a single Substitution vs some Indel
DEFAULT_MATCH_SCORE: int = 1
DEFAULT_MISMATCH_SCORE: int = -1
DEFAULT_OPEN_GAP_SCORE: int = -1
DEFAULT_EXTEND_GAP_SCORE: float = -0.1

REF_SEQUENCE_ID: str = "ref"
ALT_SEQUENCE_ID: str = "alt"

ALIGNMENT_GAP_CHAR: str = "-"

VARIANT_BASE_MATCH: str = "M"
VARIANT_BASE_MISMATCH: str = "m"
VARIANT_BASE_DELETION: str = "d"
VARIANT_BASE_INSERTION: str = "i"

HGVS_VARIANT_TYPE_SUBSTITUTION: str = "substitution"
HGVS_VARIANT_TYPE_DELETION: str = "deletion"
HGVS_VARIANT_TYPE_EXTENSION: str = "extension"
HGVS_VARIANT_TYPE_DUPLICATION: str = "duplication"
HGVS_VARIANT_TYPE_REPEAT: str = "repeat"
HGVS_VARIANT_TYPE_INSERTION: str = "insertion"
HGVS_VARIANT_TYPE_DELETION_INSERTION: str = "deletion_insertion"

MOLECULE_TYPE_ANNOTATION_KEY: str = "molecule_type"
MOLECULE_TYPE_PROTEIN: str = "protein"
HGVS_TYPE_PROTEIN: str = "p"



from sys import stderr
from unittest import TestCase
from unittest.mock import ANY, patch

from palamedes.utils import configure_logging, contains_repeated_substring, yield_repeating_substrings


class ConfigureLoggingTestCase(TestCase):
    def setUp(self):
        self.logging_patch = patch("palamedes.utils.logging")
        self.logging_mock = self.logging_patch.start()
        self.addCleanup(self.logging_patch.stop)

    def test_configure_logging(self):
        configure_logging()
        self.logging_mock.StreamHandler.assert_called_once_with(stderr)
        self.logging_mock.basicConfig.assert_called_once_with(
            level=self.logging_mock.WARNING,
            format=ANY,
            handlers=[self.logging_mock.StreamHandler()],
        )

    def test_configure_logging_debug(self):
        configure_logging(debug=True)
        self.logging_mock.StreamHandler.assert_called_once_with(stderr)
        self.logging_mock.basicConfig.assert_called_once_with(
            level=self.logging_mock.DEBUG,
            format=ANY,
            handlers=[self.logging_mock.StreamHandler()],
        )


class ContainsRepeatedSubstringTestCase(TestCase):
    def test_contains_repeated_substring_empty_string(self):
        self.assertFalse(contains_repeated_substring(""))

    def test_contains_repeated_substring_single_char(self):
        self.assertFalse(contains_repeated_substring("A"))

    def test_contains_repeated_substring_single_char_repeated(self):
        self.assertTrue(contains_repeated_substring("AA"))

    def test_contains_repeated_substring_multi_char_not_repeated(self):
        self.assertFalse(contains_repeated_substring("ATGATGCATG"))

    def test_contains_repeated_substring_multi_char_repeated(self):
        self.assertTrue(contains_repeated_substring("ATGATGATG"))


class YieldRepeatingSubstringsTestCase(TestCase):
    def test_yield_repeating_substrings_empty_string(self):
        substrings = [_ for _ in yield_repeating_substrings("")]
        self.assertEqual(substrings, [])

    def test_yield_repeating_substrings_single_char(self):
        substrings = [_ for _ in yield_repeating_substrings("A")]
        self.assertEqual(substrings, [])

    def test_yield_repeating_substrings_single_char_repeated_even(self):
        """Single char repeated 4 times, should yield the single and the double"""
        substrings = [_ for _ in yield_repeating_substrings("AAAA")]
        self.assertEqual(substrings, ["A", "AA"])

    def test_yield_repeating_substrings_single_char_repeated_odd(self):
        """Single char repeated 5 times, should yield only the single"""
        substrings = [_ for _ in yield_repeating_substrings("AAAAA")]
        self.assertEqual(substrings, ["A"])

    def test_yield_repeating_substrings_multi_char_not_repeated(self):
        """Multi character string with no repeating units"""
        substrings = [_ for _ in yield_repeating_substrings("ATGCAATGCATAGGACATGACACAC")]
        self.assertEqual(substrings, [])

    def test_yield_repeating_substrings_multi_single_repeat(self):
        """Multi character string with 1 repeating unit (ATG * 3)"""
        substrings = [_ for _ in yield_repeating_substrings("ATG" * 3)]
        self.assertEqual(substrings, ["ATG"])

    def test_yield_repeating_substrings_multi_repeat_even(self):
        """
        Multi character string with several repeating units (even): ATGATGATGATGATGATGATGATG
        - ATG * 8
        - ATGATG * 4
        - ATGATGATGATG * 2
        """
        substrings = [_ for _ in yield_repeating_substrings("ATGATG" * 4)]
        self.assertEqual(substrings, ["ATG", "ATGATG", "ATGATGATGATG"])



import logging
from functools import reduce, partial

from Bio.Align import Alignment, PairwiseAligner
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from palamedes.config import (
    ALIGNMENT_GAP_CHAR,
    DEFAULT_EXTEND_GAP_SCORE,
    DEFAULT_MATCH_SCORE,
    DEFAULT_MISMATCH_SCORE,
    DEFAULT_OPEN_GAP_SCORE,
    GLOBAL_ALIGN_MODE,
    VARIANT_BASE_DELETION,
    VARIANT_BASE_INSERTION,
    VARIANT_BASE_MATCH,
    VARIANT_BASE_MISMATCH,
    MOLECULE_TYPE_PROTEIN,
    MOLECULE_TYPE_ANNOTATION_KEY,
)
from palamedes.models import Block, VariantBlock

LOGGER = logging.getLogger(__name__)


def generate_seq_record(sequence: str, seq_id: str, molecule_type: str = MOLECULE_TYPE_PROTEIN) -> SeqRecord:
    """
    Helper function to generate a SeqRecord object from a raw input sequence. This also handles
    configuring the expected molecule_type annotation which is required for downstream steps.
    """
    return SeqRecord(
        Seq(sequence),
        id=seq_id,
        annotations={MOLECULE_TYPE_ANNOTATION_KEY: molecule_type},
    )


def reverse_seq_record(seq_record: SeqRecord) -> SeqRecord:
    """
    Helper function to copy a SeqRecord into a new one, with the sequence reversed. This is a best effort copy,
    which is only used internally to the alignment logic to more easily access the 3' end most representation of
    the best alignment.
    """
    return SeqRecord(
        Seq(seq_record.seq[::-1]),
        id=seq_record.id,
        name=seq_record.name,
        description=seq_record.description,
        dbxrefs=seq_record.dbxrefs[:],
        features=seq_record.features[:],
        annotations=seq_record.annotations.copy(),
        letter_annotations=seq_record.letter_annotations.copy(),
    )


def make_variant_base(ref_base: str, alt_base: str) -> str:
    """Helper function to generate the correct variant base given the ref and alt alignment bases"""
    if ref_base == alt_base:
        return VARIANT_BASE_MATCH

    if ref_base == ALIGNMENT_GAP_CHAR:
        return VARIANT_BASE_INSERTION

    if alt_base == ALIGNMENT_GAP_CHAR:
        return VARIANT_BASE_DELETION

    return VARIANT_BASE_MISMATCH


def can_merge_variant_blocks(left: VariantBlock, right: VariantBlock) -> bool:
    """
    Determine if 2 variant blocks can be merged, based on the following rules:
    - Must be directly adjacent (left.end == right.start)
    - No matches can appear in either
    """
    return all(
        [
            left.alignment_block.end == right.alignment_block.start,
            VARIANT_BASE_MATCH not in left.alignment_block.bases,
            VARIANT_BASE_MATCH not in right.alignment_block.bases,
        ]
    )


def merge_variant_blocks(left: VariantBlock, right: VariantBlock) -> VariantBlock:
    """
    Merge 2 variant blocks, returning the new one. This just requires that all
    sets of blocks (reference, alternate and alignment) be collapsed together
    and new reference and alt lists be created (since they can be empty).
    """
    reference_blocks = left.reference_blocks + right.reference_blocks
    if len(reference_blocks) > 0:
        new_reference_blocks = [Block.collapse(reference_blocks)]
    else:
        new_reference_blocks = []

    alternate_blocks = left.alternate_blocks + right.alternate_blocks
    if len(alternate_blocks) > 0:
        new_alternate_blocks = [Block.collapse(alternate_blocks)]
    else:
        new_alternate_blocks = []

    new_alignment_block = Block.collapse([left.alignment_block, right.alignment_block])

    return VariantBlock(new_alignment_block, new_reference_blocks, new_alternate_blocks)


def merge_reduce(
    blocks: list[VariantBlock], next_block: VariantBlock, split_consecutive_mismatches: bool = False
) -> list[VariantBlock]:
    """
    Helper function to implement merge or append, passed to functools.reduce to merge blocks. Optional
    split_consecutive_mismatches will keep consecutive mismatches from being merged, allowing for a
    chain of mismatches vs a delins. Note that this is non-standard behavior according to HGVS.
    """
    peek_block = blocks[-1]

    # base check for merge-ability
    base_can_merge = can_merge_variant_blocks(peek_block, next_block)

    # extra check for flag, merge if false OR merge if true and one of  left and right
    # are not single bp mismatches. If true and both are single base mismatches
    # then we do not merge, this 'splitting'
    should_split_mismatches = all(
        [
            split_consecutive_mismatches,
            peek_block.alignment_block.bases == VARIANT_BASE_MISMATCH,
            next_block.alignment_block.bases == VARIANT_BASE_MISMATCH,
        ]
    )

    if base_can_merge and not should_split_mismatches:
        blocks[-1] = merge_variant_blocks(peek_block, next_block)
    else:
        blocks.append(next_block)

    return blocks


def generate_alignment(
    reference_seq_record: SeqRecord,
    alternate_seq_record: SeqRecord,
    molecule_type: str = MOLECULE_TYPE_PROTEIN,
    aligner: PairwiseAligner | None = None,
) -> Alignment:
    """
    Using biopython's PairwiseAligner, generate an alignment object representing the best alignment
    between the 2 biopython SeqRecords. Note that the molecule_type argument will be used to specify a molecule_type
    annotation on the SeqRecord.

    By default the function creates an aligner object using the defaults, but the caller may provide their
    own pre-configured aligner. This aligner must be set to 'global' mode.

    Note that it is possible for multiple alignments to be returned with the same max score. Synthetic testing
    has shown that generally the highest scoring alignments are returned in "left to right" order. Take the
    following example, which was for testing HGVS duplications. This was the "intended" alignment:

        ATC---GGGGGGGG
        ATCATCGGGGGGGG

    However, there are multiple ways to get a three base insertion, all with the same score:
        ipdb> for aln in aligner.align('ATCGGGGGGGG', 'ATCATCGGGGGGGG'):
        print(aln)

        target            0 ---ATCGGGGGGGG 11
                          0 ---||||||||||| 14
        query             0 ATCATCGGGGGGGG 14

        target            0 A---TCGGGGGGGG 11
                          0 |---|||||||||| 14
        query             0 ATCATCGGGGGGGG 14

        target            0 AT---CGGGGGGGG 11
                          0 ||---||||||||| 14
        query             0 ATCATCGGGGGGGG 14

        target            0 ATC---GGGGGGGG 11
                          0 |||---|||||||| 14
        query             0 ATCATCGGGGGGGG 14

    Due to the HGVS rule of always indicating that the 3' most base is considered the modified one, a best effort
    attempt is made to return the most "right aligned" alignment, by returning the last alignment with the highest
    score. This way not yield the ideal results in more complicated cases.
    """
    if aligner is not None:
        if aligner.mode != GLOBAL_ALIGN_MODE:
            raise ValueError(f"Custom PairwiseAligner must be set to global mode, got: {aligner.mode}")
    else:
        aligner = PairwiseAligner(
            mode=GLOBAL_ALIGN_MODE,
            match_score=DEFAULT_MATCH_SCORE,
            mismatch_score=DEFAULT_MISMATCH_SCORE,
            open_gap_score=DEFAULT_OPEN_GAP_SCORE,
            extend_gap_score=DEFAULT_EXTEND_GAP_SCORE,
        )

    if (ref_molecule_type := reference_seq_record.annotations.get("molecule_type")) != molecule_type:
        raise ValueError(
            "Cannot generate alignment, reference_seq_record is a SeqRecord an invalid molecule_type annotation "
            f"got: {ref_molecule_type}, expected: {molecule_type}!"
        )

    if (alt_molecule_type := alternate_seq_record.annotations.get("molecule_type")) != molecule_type:
        raise ValueError(
            "Cannot generate alignment, alternate_seq_record is a SeqRecord an invalid molecule_type annotation "
            f"got: {alt_molecule_type}, expected: {molecule_type}!"
        )

    # reverse the sequences and then align the reversed, keeping the first best alignment
    reversed_ref_seq_record = reverse_seq_record(reference_seq_record)
    reversed_alt_seq_record = reverse_seq_record(alternate_seq_record)
    reversed_alignments = aligner.align(reversed_ref_seq_record, reversed_alt_seq_record)
    reversed_alignment = reversed_alignments[0]

    # undo the reversal, to recover the "last" highest scoring alignment for the forward
    # which should correspond to the 3' end most alignment and follow HGSV spec
    # note 'infer_coordinates' will soon be deprecated, update to 'parse_printed_alignment' once its live
    forward_coordinates = Alignment.infer_coordinates(
        [
            reversed_alignment[0][::-1],
            reversed_alignment[1][::-1],
        ]
    )
    forward_alignment = Alignment(
        [reference_seq_record, alternate_seq_record],
        forward_coordinates,
    )

    # score is not technically an attribute on the class
    setattr(forward_alignment, "score", reversed_alignment.score)

    return forward_alignment


def generate_variant_blocks(alignment: Alignment, split_consecutive_mismatches: bool = False) -> list[VariantBlock]:
    """
    Given a BioPython.Alignment object, parse the alignment to generate a list of VariantBlock objects.
    A VariantBlock is an internal object which represents a contiguous run of positions within the alignment
    which are not matches (mismatch, del or ins). These blocks will be categorized and converted into HGVS
    objects in a later step, but this intermediate representation is useful for debugging and testing.

    This function uses functools.reduce in order to structure the problem of merging blocks
    into a more understandable approach:
    - We define the logic for when 2 VariantBlocks can be merged together, see can_merge_variant_blocks
    - We define the logic for how to merge 2 VariantBlocks together, see merge_variant_blocks
    - We define a reduce function that treats the accumulator like a stack, peeking the head, checking
      for merge-ability and merging if possible at each iteration.
    """
    reference_indices = alignment.indices[0].tolist()
    alternate_indices = alignment.indices[1].tolist()
    bases_zipper = zip(alignment[0], alignment[1])

    # First generate a VariantBlock for all positions in the Alignment along with the relevant sequence Blocks
    single_position_variant_blocks = [
        VariantBlock(
            Block(idx, idx + 1, make_variant_base(reference_base, alternate_base)),
            []
            if reference_base == ALIGNMENT_GAP_CHAR
            else [
                Block(
                    reference_indices[idx],
                    reference_indices[idx] + 1,
                    reference_base,
                )
            ],
            []
            if alternate_base == ALIGNMENT_GAP_CHAR
            else [
                Block(
                    alternate_indices[idx],
                    alternate_indices[idx] + 1,
                    alternate_base,
                )
            ],
        )
        for idx, (reference_base, alternate_base) in enumerate(bases_zipper)
    ]

    # use reduce with a helper function to merge valid/adjacent blocks together
    wrapped_merge_func = partial(merge_reduce, split_consecutive_mismatches=split_consecutive_mismatches)
    merged_blocks = reduce(wrapped_merge_func, single_position_variant_blocks[1:], [single_position_variant_blocks[0]])

    # filter out the matches
    return [
        merged_block for merged_block in merged_blocks if VARIANT_BASE_MATCH not in merged_block.alignment_block.bases
    ]


def get_upstream_reference_sequence(alignment: Alignment, anchor_position: int, num_bases: int) -> str:
    """
    Get num_bases of the reference sequence directly upstream of a variant block. This is done by treating
    the starting position of the variant block as an anchor into the aligned/gapped reference sequence.
    A slice is then taken from 0 -> the anchor position, and gaps are removed from this slice. Finally,
    return the last num_bases of the substring. This is done to avoid cases where an upstream gap
    may cause a more naive approach to miss a duplication or repeat (since a gap might appear in the checked sequence).
    """
    ungapped_ref = alignment[0][:anchor_position].replace(ALIGNMENT_GAP_CHAR, "")
    return ungapped_ref[-num_bases:]



from Bio.Align import Alignment
from hgvs.edit import (
    Repeat,
    Dup,
    AAExt,
    AARefAlt,
    AASub,
)
from hgvs.posedit import PosEdit
from hgvs.location import Interval, AAPosition
from hgvs.sequencevariant import SequenceVariant

from palamedes.align import get_upstream_reference_sequence
from palamedes.models import VariantBlock
from palamedes.config import (
    HGVS_VARIANT_TYPE_SUBSTITUTION,
    HGVS_VARIANT_TYPE_DELETION,
    HGVS_VARIANT_TYPE_EXTENSION,
    HGVS_VARIANT_TYPE_DUPLICATION,
    HGVS_VARIANT_TYPE_REPEAT,
    HGVS_VARIANT_TYPE_INSERTION,
    HGVS_VARIANT_TYPE_DELETION_INSERTION,
    HGVS_TYPE_PROTEIN,
    MOLECULE_TYPE_PROTEIN,
)
from palamedes.utils import yield_repeating_substrings, zbho_to_obfc, zb_to_ob, zb_position_to_end_coordinate


class HgvsProteinBuilder:
    def __init__(self, alignment: Alignment) -> None:
        self._alignment = alignment

    def build(self, variant_block: VariantBlock, hgvs_type: str) -> SequenceVariant:
        pos_edit_builder_funcs = {
            HGVS_VARIANT_TYPE_SUBSTITUTION: self._build_substitution,
            HGVS_VARIANT_TYPE_DELETION: self._build_deletion,
            HGVS_VARIANT_TYPE_INSERTION: self._build_insertion,
            HGVS_VARIANT_TYPE_EXTENSION: self._build_extension,
            HGVS_VARIANT_TYPE_DUPLICATION: self._build_duplication,
            HGVS_VARIANT_TYPE_REPEAT: self._build_repeat,
            HGVS_VARIANT_TYPE_DELETION_INSERTION: self._build_deletion_insertion,
        }
        pos_edit = pos_edit_builder_funcs[hgvs_type](variant_block)

        return SequenceVariant(ac=self._alignment.target.id, type=HGVS_TYPE_PROTEIN, posedit=pos_edit)

    def _build_substitution(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein substitution build logic, this is a fairly simple case since the variant block has ref + alt data and
        the variant is always a single position. Convert the coordinates to OBFC and return the PosEdit.
        """
        reference_block = variant_block.reference_blocks[0]
        alternate_block = variant_block.alternate_blocks[0]
        start_obfc, end_obfc = zbho_to_obfc(reference_block.start, reference_block.end)

        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=start_obfc,
                    aa=reference_block.bases,
                ),
                end=AAPosition(
                    base=end_obfc,
                    aa=reference_block.bases,
                ),
            ),
            edit=AASub(
                ref=reference_block.bases,
                alt=alternate_block.bases,
            ),
        )

    def _build_deletion(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein deletion build logic, this is a fairly simple case since the variant block has the ref data,
        including the positions + sequence that was deleted. Convert the coordinates to OBFC and return the PosEdit.
        """
        reference_block = variant_block.reference_blocks[0]
        start_obfc, end_obfc = zbho_to_obfc(reference_block.start, reference_block.end)

        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=start_obfc,
                    aa=reference_block.bases[0],
                ),
                end=AAPosition(
                    base=end_obfc,
                    aa=reference_block.bases[-1],
                ),
            ),
            edit=AARefAlt(
                ref=reference_block.bases,
                alt=None,
            ),
        )

    def _build_insertion(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein insertion build logic, this is a more complicated example since the ref data does not exist on the
        variant block and has to be looked up from the Alignment. We leverage the .indices field on the Alignment
        which maps alignment coordinates back to sequence coordinates (gaps have -1). This is a numpy array so
        all values are cast to python integers before being saved into an HGVS object.

        The Interval we want is: (1 base upstream of the insert, 1 base downstream of the insert). To get there we
        do the following:
        - Get the start position of the insert in the alignment. This start is the index for that position within the
          .indices array. Subtract 1 from that to get the index of the next base upstream (the flanking ref base). This
          could out of bounds if the variant has not been categorized properly (should be an extension in this case).
        - Lookup the ref start index from .indices using the computed index above
        - Lookup the ref end index from .indices using the end position of the insert in the alignment. No offset is
          needed here since the end of a ZBHO interval is the same as the index for the next base.
        - Convert the 2 zero based indices into one based indices.
        - Build and return the object, using the zero based indices to get the anchor bases from the ref sequence
        """
        upstream_ref_base_index = variant_block.alignment_block.start - 1
        ref_flanking_start_position = int(self._alignment.indices[0][upstream_ref_base_index])
        ref_flanking_end_position = int(self._alignment.indices[0][variant_block.alignment_block.end])

        ref_flanking_start_position_ob = zb_to_ob(ref_flanking_start_position)
        ref_flanking_end_position_ob = zb_to_ob(ref_flanking_end_position)

        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=ref_flanking_start_position_ob,
                    aa=self._alignment.target.seq[ref_flanking_start_position],
                ),
                end=AAPosition(
                    base=ref_flanking_end_position_ob,
                    aa=self._alignment.target.seq[ref_flanking_end_position],
                ),
            ),
            edit=AARefAlt(
                ref=None,
                alt=variant_block.alternate_blocks[0].bases,
            ),
        )

    def _build_extension(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein extension build logic. This case is simple since the Interval is always the first or the last base.
        Check if it is a start or end extension and get the first or last reference base from the ref sequence. The
        position is either 1 or the last base (len of the ref sequence). Since the values are static, they are defined
        in OB already.
        """
        is_start = variant_block.alignment_block.start == 0
        ref_base = self._alignment.target.seq[0] if is_start else self._alignment.target.seq[-1]
        ref_position = 1 if is_start else len(self._alignment.target.seq)

        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=ref_position,
                    aa=ref_base,
                ),
                end=AAPosition(
                    base=ref_position,
                    aa=ref_base,
                ),
            ),
            edit=AAExt(
                ref=ref_base,
                aaterm=variant_block.alternate_blocks[0].bases,
                length=len(variant_block.alternate_blocks[0].bases) * (-1 if is_start else 1),
            ),
        )

    def _build_duplication(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein duplication build logic. This is another complicated case that requires some index munging with the
        reference sequence. It is similar to the insertion except that we want the duplicated bases/positions, not
        the bases/positions that flank the insertion. So we get the index for the first upstream base and convert it
        to an end coordinate. We then subtract the length of the insertion from the end to get the start. Finally,
        convert to OBFC. The edit object is just an empty Dup().
        """
        upstream_ref_base_index = variant_block.alignment_block.start - 1
        ref_duplication_end_position = zb_position_to_end_coordinate(
            self._alignment.indices[0][upstream_ref_base_index]
        )
        ref_duplication_start_position = ref_duplication_end_position - len(variant_block.alternate_blocks[0].bases)
        ref_duplication_start_position_obfc, ref_duplication_end_position_obfc = zbho_to_obfc(
            ref_duplication_start_position, ref_duplication_end_position
        )

        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=ref_duplication_start_position_obfc,
                    aa=variant_block.alternate_blocks[0].bases[0],
                ),
                end=AAPosition(
                    base=ref_duplication_end_position_obfc,
                    aa=variant_block.alternate_blocks[0].bases[-1],
                ),
            ),
            edit=Dup(),
        )

    def _build_repeat(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein repeat build logic. This is the most complicated case, and it builds on the duplication case. The
        key difference is that instead of taking the insert sequence as the matching upstream element, we first
        have to find the largest repeating sub-string that matches the directly upstream bases. Once that is found,
        the interval covering the reference bases being repeated are found using the same logic as with duplication, but
        using the sub-string for the length. The last thing is to find the repeat number, which can be computed based on
        integer division between the length of the insert and the length of the largest repeat.
        """
        largest_upstream_repeat = [
            substring
            for substring in yield_repeating_substrings(variant_block.alternate_blocks[0].bases)
            if get_upstream_reference_sequence(self._alignment, variant_block.alignment_block.start, len(substring))
            == substring
        ][-1]

        upstream_ref_base_index = variant_block.alignment_block.start - 1
        ref_duplication_end_position = zb_position_to_end_coordinate(
            self._alignment.indices[0][upstream_ref_base_index]
        )
        ref_duplication_start_position = ref_duplication_end_position - len(largest_upstream_repeat)
        start_obfc, end_obfc = zbho_to_obfc(ref_duplication_start_position, ref_duplication_end_position)

        repeat_value = len(variant_block.alternate_blocks[0].bases) // len(largest_upstream_repeat)

        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=start_obfc,
                    aa=largest_upstream_repeat[0],
                ),
                end=AAPosition(
                    base=end_obfc,
                    aa=largest_upstream_repeat[-1],
                ),
            ),
            edit=Repeat(min=repeat_value, max=repeat_value),
        )

    def _build_deletion_insertion(self, variant_block: VariantBlock) -> PosEdit:
        """
        Protein delins logic. Fairly simple logic by taking the ref + alt blocks and combining them together.
        """
        reference_block = variant_block.reference_blocks[0]
        alternate_block = variant_block.alternate_blocks[0]

        start_obfc, end_obfc = zbho_to_obfc(reference_block.start, reference_block.end)
        return PosEdit(
            pos=Interval(
                start=AAPosition(
                    base=start_obfc,
                    aa=reference_block.bases[0],
                ),
                end=AAPosition(
                    base=end_obfc,
                    aa=reference_block.bases[-1],
                ),
            ),
            edit=AARefAlt(
                ref=reference_block.bases,
                alt=alternate_block.bases,
            ),
        )


BUILDER_CONFIG = {
    MOLECULE_TYPE_PROTEIN: HgvsProteinBuilder,
}



import logging
import sys
from typing import Iterable


def configure_logging(debug: bool = False) -> None:
    stderr_handler = logging.StreamHandler(sys.stderr)
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.WARNING,
        format="[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        handlers=[stderr_handler],
    )


def contains_repeated_substring(input_string: str) -> bool:
    """
    Common leetcode problem to determine if an input_string has a repeating sub-string which
    can be repeated 1 or more times to recreate the input string. This leverages the fact
    that such a string is a non-trivial rotation of itself and a string A is a rotation of
    another string B if A is a sub-string of B+B.

    We use find on B+B to see the location of the sub-string. It cannot be -1 (not found) or
    the length of the input sequence (since that is just the double).

    This is adapted from:
    https://stackoverflow.com/questions/55823298/how-do-i-check-if-a-string-is-entirely-made-of-the-same-substring
    """
    doubled = input_string + input_string
    return doubled.find(input_string, 1) not in {-1, len(input_string)}


def yield_repeating_substrings(input_string: str) -> Iterable[str]:
    """
    Generator to yield sub-strings of the input sequence which can be repeated some number of times
    to regenerate the input sequence. This is similar in nature to contains_repeated_substring, but
    the actual sub-strings are returned. The logic is like so:
    - Start with a length 1 sub-string (the first character)
    - Check if 1 (the length) is a divisor of the input string length (remainder of divmod == 0)
    - If so, check if the sub-string repeated q times (the quotient) equals the input_string (and yield if so)
    - Repeat the above with larger and larger sub-strings (starting from zero, but adding a new character each time)
    - Stop when you sub-string doubled in length would be larger then the input sequence
    """
    input_len = len(input_string)
    idx = 0

    while ((idx + 1) * 2) <= input_len:
        substr_for_idx = input_string[0 : idx + 1]
        quotient, remainder = divmod(input_len, len(substr_for_idx))
        if remainder == 0 and substr_for_idx * quotient == input_string:
            yield substr_for_idx

        idx += 1


def zb_to_ob(zb: int) -> int:
    """
    Convert a single zero based position (or sequence index) into one based. This means simply adding 1 to it.
    This function simply exists to make things more explicit and avoid having magic "+ 1"s everywhere.
    """
    return zb + 1


def zbho_to_obfc(zbho_start: int, zbho_end: int) -> tuple[int, int]:
    """
    Convert a zero based half open (start, end) into a one based fully closed, based on the following:
    - add 1 to both positions to account for zero ==> one
    - subtract 1 to the end position to account for half open to fully closed
    """
    return (
        zb_to_ob(zbho_start),
        zb_to_ob(zbho_end) - 1,
    )


def zb_position_to_end_coordinate(zb_position: int) -> int:
    """
    Convert a zero based position (or index) into the end coordinate of a ZBHO interval including
    this position as the ending position. This is done by simply adding 1 to the position to make it
    half open. Again this is just to make this clearer and more explicit.
    """
    return zb_position + 1



from Bio.Align import Alignment, PairwiseAligner

from palamedes.align import (
    generate_alignment,
    make_variant_base,
    can_merge_variant_blocks,
    merge_variant_blocks,
    generate_seq_record,
    generate_variant_blocks,
)
from palamedes.config import (
    GLOBAL_ALIGN_MODE,
    REF_SEQUENCE_ID,
    VARIANT_BASE_MATCH,
    ALIGNMENT_GAP_CHAR,
    VARIANT_BASE_INSERTION,
    VARIANT_BASE_DELETION,
    VARIANT_BASE_MISMATCH,
    MOLECULE_TYPE_PROTEIN,
    MOLECULE_TYPE_ANNOTATION_KEY,
)
from palamedes.models import VariantBlock, Block
from tests.base import PalamedesBaseCase


class GenerateSeqRecordsTestCase(PalamedesBaseCase):
    def test_generate_seq_record(self):
        seq = "AAA"
        seq_rec = generate_seq_record(seq, REF_SEQUENCE_ID)

        self.assertEqual(seq_rec.seq, seq)
        self.assertEqual(seq_rec.id, REF_SEQUENCE_ID)
        self.assertEqual(seq_rec.annotations, {MOLECULE_TYPE_ANNOTATION_KEY: MOLECULE_TYPE_PROTEIN})

    def test_generate_seq_record_custom_molecule_type(self):
        seq = "AAA"
        custom_molecule_type = "dna"

        seq_rec = generate_seq_record(seq, REF_SEQUENCE_ID, molecule_type=custom_molecule_type)
        self.assertEqual(seq_rec.annotations, {MOLECULE_TYPE_ANNOTATION_KEY: custom_molecule_type})


class MakeVariantBaseTestCase(PalamedesBaseCase):
    def test_make_variant_base_match(self):
        self.assertEqual(make_variant_base("A", "A"), VARIANT_BASE_MATCH)

    def test_make_variant_base_mismatch(self):
        self.assertEqual(make_variant_base("A", "T"), VARIANT_BASE_MISMATCH)

    def test_make_variant_base_deletion(self):
        self.assertEqual(make_variant_base("A", ALIGNMENT_GAP_CHAR), VARIANT_BASE_DELETION)

    def test_make_variant_base_insertion(self):
        self.assertEqual(make_variant_base(ALIGNMENT_GAP_CHAR, "T"), VARIANT_BASE_INSERTION)


class CanMergeVariantBlocksTestCase(PalamedesBaseCase):
    def make_variant_block(self, start: int, end: int, bases: str) -> VariantBlock:
        return VariantBlock(Block(start, end, bases), [], [])

    def test_can_merge_variant_blocks_not_adjacent(self):
        left = self.make_variant_block(0, 4, VARIANT_BASE_DELETION * 4)
        right = self.make_variant_block(6, 8, VARIANT_BASE_MISMATCH * 2)
        self.assertFalse(can_merge_variant_blocks(left, right))

    def test_can_merge_variant_blocks_left_has_match(self):
        left = self.make_variant_block(0, 4, VARIANT_BASE_DELETION * 3 + VARIANT_BASE_MATCH)
        right = self.make_variant_block(4, 6, VARIANT_BASE_MISMATCH * 2)
        self.assertFalse(can_merge_variant_blocks(left, right))

    def test_can_merge_variant_blocks_right_has_match(self):
        left = self.make_variant_block(0, 4, VARIANT_BASE_DELETION * 3 + VARIANT_BASE_MISMATCH)
        right = self.make_variant_block(4, 6, VARIANT_BASE_MISMATCH + VARIANT_BASE_MATCH)
        self.assertFalse(can_merge_variant_blocks(left, right))

    def test_can_merge_variant_blocks_both_have_match(self):
        left = self.make_variant_block(0, 4, VARIANT_BASE_DELETION * 3 + VARIANT_BASE_MATCH)
        right = self.make_variant_block(4, 6, VARIANT_BASE_MISMATCH + VARIANT_BASE_MATCH)
        self.assertFalse(can_merge_variant_blocks(left, right))

    def test_can_merge_variant_blocks_can_merge(self):
        left = self.make_variant_block(0, 4, VARIANT_BASE_DELETION * 3 + VARIANT_BASE_MISMATCH)
        right = self.make_variant_block(4, 6, VARIANT_BASE_MISMATCH + VARIANT_BASE_INSERTION)
        self.assertTrue(can_merge_variant_blocks(left, right))


class MergeVariantBlocksTestCase(PalamedesBaseCase):
    def test_merge_variant_blocks_simple(self):
        """Test merging 2 single base mismatches"""
        left = VariantBlock(
            Block(0, 1, VARIANT_BASE_MISMATCH),
            [Block(0, 1, "A")],
            [Block(0, 1, "T")],
        )
        right = VariantBlock(
            Block(1, 2, VARIANT_BASE_MISMATCH),
            [Block(1, 2, "G")],
            [Block(1, 2, "C")],
        )

        merged = merge_variant_blocks(left, right)

        self.assertEqual(merged.alignment_block.start, left.alignment_block.start)
        self.assertEqual(merged.alignment_block.end, right.alignment_block.end)
        self.assertEqual(merged.alignment_block.bases, VARIANT_BASE_MISMATCH + VARIANT_BASE_MISMATCH)

        self.assertEqual(len(merged.reference_blocks), 1)
        self.assertEqual(merged.reference_blocks[0].start, left.reference_blocks[0].start)
        self.assertEqual(merged.reference_blocks[0].end, right.reference_blocks[0].end)
        self.assertEqual(
            merged.reference_blocks[0].bases, left.reference_blocks[0].bases + right.reference_blocks[0].bases
        )

        self.assertEqual(len(merged.alternate_blocks), 1)
        self.assertEqual(merged.alternate_blocks[0].start, left.alternate_blocks[0].start)
        self.assertEqual(merged.alternate_blocks[0].end, right.alternate_blocks[0].end)
        self.assertEqual(
            merged.alternate_blocks[0].bases, left.alternate_blocks[0].bases + right.alternate_blocks[0].bases
        )

    def test_merge_variant_blocks_complex(self):
        """Test merging 2 complex VariantBlocks based on the following hypothetical alignment:
         T T T -
         - - C G
        0 1 2 3 4 GLOBAL
        0 1 2 3 . REF
        . . 0 1 2 ALT
        The left side will already have ddm, we are merging in the insertion for the test.
        """
        left = VariantBlock(
            Block(0, 3, "".join([VARIANT_BASE_DELETION, VARIANT_BASE_DELETION, VARIANT_BASE_MISMATCH])),
            [Block(0, 3, "TTT")],
            [Block(0, 1, "C")],
        )
        right = VariantBlock(
            Block(3, 4, VARIANT_BASE_INSERTION),
            [],
            [Block(1, 2, "G")],
        )

        merged = merge_variant_blocks(left, right)

        self.assertEqual(merged.alignment_block.start, left.alignment_block.start)
        self.assertEqual(merged.alignment_block.end, right.alignment_block.end)
        self.assertEqual(
            merged.alignment_block.bases,
            "".join([VARIANT_BASE_DELETION, VARIANT_BASE_DELETION, VARIANT_BASE_MISMATCH, VARIANT_BASE_INSERTION]),
        )

        # there is no right ref block, so everything is on the left one
        self.assertEqual(len(merged.reference_blocks), 1)
        self.assertEqual(merged.reference_blocks[0], left.reference_blocks[0])

        self.assertEqual(len(merged.alternate_blocks), 1)
        self.assertEqual(merged.alternate_blocks[0].start, left.alternate_blocks[0].start)
        self.assertEqual(merged.alternate_blocks[0].end, right.alternate_blocks[0].end)
        self.assertEqual(
            merged.alternate_blocks[0].bases, left.alternate_blocks[0].bases + right.alternate_blocks[0].bases
        )


class GenerateAlignmentTestCase(PalamedesBaseCase):
    def test_generate_alignment_missing_molecule_type_error(self):
        ref, alt = self.make_seq_records("A", "A", molecule_type="foobar")
        del ref.annotations[MOLECULE_TYPE_ANNOTATION_KEY]
        with self.assertRaisesRegex(ValueError, "got: None"):
            generate_alignment(ref, alt)

    def test_generate_alignment_wrong_molecule_type_error(self):
        ref, alt = self.make_seq_records("A", "A", molecule_type="foobar")

        with self.assertRaisesRegex(ValueError, f"expected: {MOLECULE_TYPE_PROTEIN}"):
            generate_alignment(ref, alt)

    def test_generate_alignment_custom_aligner_mode_error(self):
        local_mode = "local"
        custom_aligner = PairwiseAligner(mode=local_mode)
        ref, alt = self.make_seq_records("A", "T")
        with self.assertRaisesRegex(ValueError, f"got: {local_mode}"):
            generate_alignment(ref, alt, aligner=custom_aligner)

    def test_generate_alignment(self):
        ref, alt = self.make_seq_records("A", "T")

        alignment = generate_alignment(ref, alt)

        self.assertTrue(isinstance(alignment, Alignment))
        self.assertIs(alignment.target, ref)
        self.assertIs(alignment.query, alt)

    def test_generate_alignment_custom_aligner(self):
        ref, alt = self.make_seq_records("A", "A")

        custom_match_score = 10_000
        custom_aligner = PairwiseAligner(mode=GLOBAL_ALIGN_MODE, match_score=custom_match_score)

        alignment = generate_alignment(ref, alt, aligner=custom_aligner)
        self.assertTrue(alignment.score, custom_match_score)

    def test_generate_alignment_three_prime_end_most_del(self):
        ref, alt = self.make_seq_records(
            "T" + "A" * 10 + "G",
            "T" + "A" * 9 + "G",
        )

        alignment = generate_alignment(ref, alt)

        # ensure we get the ref unchanged
        # and the last A deleted, not any others
        self.assertEqual(alignment[0], ref.seq)
        self.assertEqual(alignment[1], "T" + "A" * 9 + "-G")

    def test_generate_alignment_three_prime_end_most_double_del(self):
        ref, alt = self.make_seq_records(
            "ATTGCCA",
            "ATGCA",
        )

        alignment = generate_alignment(ref, alt)

        # ensure we get the ref unchanged
        # and the second of each paired based deleted
        self.assertEqual(alignment[0], ref.seq)
        self.assertEqual(alignment[1], "AT-GC-A")

    def test_generate_alignment_three_prime_end_most_ins(self):
        ref, alt = self.make_seq_records(
            "T" + "A" * 9 + "G",
            "T" + "A" * 10 + "G",
        )

        alignment = generate_alignment(ref, alt)

        # ensure we get the alt unchanged
        # and the insertion happens after the last A in the ref
        self.assertEqual(alignment[0], "T" + "A" * 9 + "-G")
        self.assertEqual(alignment[1], alt.seq)

    def test_generate_alignment_three_prime_end_most_double_ins(self):
        ref, alt = self.make_seq_records(
            "ATGCA",
            "ATTGCCA",
        )
        alignment = generate_alignment(ref, alt)

        # ensure we get the alt unchanged
        # and the insertion happens after the last A in the ref
        self.assertEqual(alignment[0], "AT-GC-A")
        self.assertEqual(alignment[1], alt.seq)


class GenerateVariantBlocksTestCase(PalamedesBaseCase):
    def test_generate_variant_blocks_all_matches(self):
        alignment = self.make_alignment("A" * 5, "A" * 5)
        self.assertEqual(generate_variant_blocks(alignment), [])

    def test_generate_variant_blocks_all_matches_single_mismatch(self):
        alignment = self.make_alignment("ACT", "AGT")
        variant_blocks = generate_variant_blocks(alignment)

        self.assertEqual(len(variant_blocks), 1)
        self.assertEqual(variant_blocks[0].alignment_block, Block(1, 2, VARIANT_BASE_MISMATCH))
        self.assertEqual(variant_blocks[0].reference_blocks, [Block(1, 2, "C")])
        self.assertEqual(variant_blocks[0].alternate_blocks, [Block(1, 2, "G")])

    def test_generate_variant_blocks_merging(self):
        """
        Test an alignment with a 3 position variant in order, deletion -> mismatch -> insertion
            A T C - T
            A - G A T
           0 1 2 3 4 5 GLOBAL
           0 1 2 3 . 4 REF
           0 . 1 2 3 4 ALT
        """
        alignment = self.make_alignment(
            "ATC-T",
            "A-GAT",
        )
        variant_blocks = generate_variant_blocks(alignment)

        self.assertEqual(len(variant_blocks), 1)
        self.assertEqual(
            variant_blocks[0].alignment_block,
            Block(1, 4, "".join([VARIANT_BASE_DELETION, VARIANT_BASE_MISMATCH, VARIANT_BASE_INSERTION])),
        )
        self.assertEqual(variant_blocks[0].reference_blocks, [Block(1, 3, "TC")])
        self.assertEqual(variant_blocks[0].alternate_blocks, [Block(1, 3, "GA")])

    def test_generate_variant_blocks_complex(self):
        """
        Test an alignment with a number of variants, some merged and some not
            A T C T - - T
            A - C G A A T
           0 1 2 3 4 5 6 7 GLOBAL
           0 1 2 3 4 . . 6 REF
           0 . 1 2 3 4 5 6 ALT

        The full "variant bases" would be: MdMmiiM
        """
        alignment = self.make_alignment(
            "ATCT--T",
            "A-CGAAT",
        )
        variant_blocks = generate_variant_blocks(alignment)

        # mii gets merged so 2 total
        self.assertEqual(len(variant_blocks), 2)
        self.assertEqual(
            variant_blocks[0].alignment_block,
            Block(1, 2, VARIANT_BASE_DELETION),
        )
        self.assertEqual(variant_blocks[0].reference_blocks, [Block(1, 2, "T")])
        self.assertEqual(variant_blocks[0].alternate_blocks, [])

        self.assertEqual(
            variant_blocks[1].alignment_block,
            Block(3, 6, "".join([VARIANT_BASE_MISMATCH, VARIANT_BASE_INSERTION, VARIANT_BASE_INSERTION])),
        )
        self.assertEqual(variant_blocks[1].reference_blocks, [Block(3, 4, "T")])
        self.assertEqual(variant_blocks[1].alternate_blocks, [Block(2, 5, "GAA")])

    def test_generate_variant_blocks_split_subs(self):
        alignment = self.make_alignment(
            "AAAA",
            "TTTT",
        )
        variant_blocks = generate_variant_blocks(alignment, split_consecutive_mismatches=True)

        self.assertEqual(len(variant_blocks), 4)
        for idx, block in enumerate(variant_blocks):
            expected_start = idx
            expecte_end = idx + 1
            self.assertEqual(block.alignment_block, Block(expected_start, expecte_end, VARIANT_BASE_MISMATCH))
            self.assertEqual(block.reference_blocks, [Block(expected_start, expecte_end, "A")])
            self.assertEqual(block.alternate_blocks, [Block(expected_start, expecte_end, "T")])

    def test_generate_variant_blocks_no_split_indels(self):
        """Test split flag does not change how actual indels are treated"""
        alignment = self.make_alignment(
            "AAAA---",
            "---TTTT",
        )
        variant_blocks = generate_variant_blocks(alignment, split_consecutive_mismatches=True)
        self.assertEqual(len(variant_blocks), 1)

        self.assertEqual(
            variant_blocks[0].alignment_block,
            Block(
                0,
                7,
                "".join(
                    [
                        VARIANT_BASE_DELETION,
                        VARIANT_BASE_DELETION,
                        VARIANT_BASE_DELETION,
                        VARIANT_BASE_MISMATCH,
                        VARIANT_BASE_INSERTION,
                        VARIANT_BASE_INSERTION,
                        VARIANT_BASE_INSERTION,
                    ]
                ),
            ),
        )



from unittest import TestCase

from Bio.Align import Alignment
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from palamedes.config import (
    MOLECULE_TYPE_PROTEIN,
    MOLECULE_TYPE_ANNOTATION_KEY,
    REF_SEQUENCE_ID,
    ALT_SEQUENCE_ID,
    ALIGNMENT_GAP_CHAR,
)


class PalamedesBaseCase(TestCase):
    def make_seq_records(
        self, ref_seq: str, alt_seq: str, molecule_type: str = MOLECULE_TYPE_PROTEIN
    ) -> tuple[SeqRecord, SeqRecord]:
        ref = SeqRecord(Seq(ref_seq), id=REF_SEQUENCE_ID, annotations={MOLECULE_TYPE_ANNOTATION_KEY: molecule_type})
        alt = SeqRecord(Seq(alt_seq), id=ALT_SEQUENCE_ID, annotations={MOLECULE_TYPE_ANNOTATION_KEY: molecule_type})
        return ref, alt

    def make_alignment(self, ref_aligned_bases: str, alt_aligned_bases: str) -> Alignment:
        coords = Alignment.infer_coordinates([ref_aligned_bases, alt_aligned_bases])
        ref, alt = self.make_seq_records(
            ref_aligned_bases.replace(ALIGNMENT_GAP_CHAR, ""), alt_aligned_bases.replace(ALIGNMENT_GAP_CHAR, "")
        )
        return Alignment([ref, alt], coords)



from Bio.Align import PairwiseAligner
from Bio.SeqRecord import SeqRecord
from hgvs.sequencevariant import SequenceVariant

from palamedes.config import MOLECULE_TYPE_PROTEIN, ALT_SEQUENCE_ID, REF_SEQUENCE_ID
from palamedes.align import generate_seq_record, generate_alignment, generate_variant_blocks
from palamedes.hgvs.utils import categorize_variant_block
from palamedes.hgvs.builders import BUILDER_CONFIG


__version__ = "0.0.7"


def generate_hgvs_variants(
    reference_sequence: str | SeqRecord,
    alternate_sequence: str | SeqRecord,
    molecule_type: str = MOLECULE_TYPE_PROTEIN,
    aligner: PairwiseAligner | None = None,
    use_non_standard_substitution_rules: bool = False,
) -> list[SequenceVariant]:
    """
    Given the reference and alternate sequences, as either strings or `Bio.SeqRecord.SeqRecord` objects, compute the
    alignment, find the variants and build `hgvs.sequencevariant.SequenceVariant` objects from them. Returning the list
    of objects at the end. Note the following things:

    - Variants and their HGVS objects are based on consecutive alignment positions that are non matches, any match will split or disrupt the variant.
    - Variant and their HGVS objects are handled in isolation from others, they do not "account" for things that happen upstream of their position in the alignment.

    A custom `PairwiseAligner` may be provided to tweak the alignment score, but the default one uses these scores:

    - mode: `global` (this is required, even for a custom one)
    - match score: `1`
    - mismatch score: `-1`
    - open gap score: `-1`
    - extend gap score: `-0.1`

    An optional flag: `use_non_standard_substitution_rules` can be passed as `True`, which will enable logic that treats
    multiple consecutive mismatches as seperare subsitutions, vs merging together into a delins. This is against HGVS
    spec but has utility for some use cases.

    If using pre-built `SeqRecord` objects, be sure to set the `molecule_type` annotation key to a supported molecule type
    and pass in the corresponding molecule_type as an input. At the time of writing, only `protein` is supported as a
    molecule type.

    Example using a raw string:

    .. code-block:: python

        >>> from palamedes import generate_hgvs_variants
        >>> generate_hgvs_variants("PFKISIHL", "TPFKISIH")
        [
            SequenceVariant(ac=ref, type=p, posedit=Pro1extThr-1, gene=None),
            SequenceVariant(ac=ref, type=p, posedit=Leu8del, gene=None),
        ]

    Example using pre-configured `SeqRecord` objects:

    .. code-block:: python

        >>> from Bio.Seq import Seq
        >>> from Bio.SeqRecord import SeqRecord
        >>> from palamedes import generate_hgvs_variants
        >>> ref = SeqRecord(Seq("PFKISIHL"), id="Jelleine-I", annotations={"molecule_type": "protein"})
        >>> alt = SeqRecord(Seq("TPFKISIH"), id="Jelleine-IV", annotations={"molecule_type": "protein"})
        >>> generate_hgvs_variants(ref, alt)
        [
            SequenceVariant(ac=Jelleine-I, type=p, posedit=Pro1extThr-1, gene=None),
            SequenceVariant(ac=Jelleine-I, type=p, posedit=Leu8del, gene=None)
        ]
    """

    if molecule_type not in BUILDER_CONFIG:
        raise NotImplementedError(f"No HGVS builder is defined for molecule_type: {molecule_type}!")

    ref_seq_record = (
        generate_seq_record(reference_sequence, REF_SEQUENCE_ID, molecule_type=molecule_type)
        if isinstance(reference_sequence, str)
        else reference_sequence
    )

    alt_seq_record = (
        generate_seq_record(alternate_sequence, ALT_SEQUENCE_ID, molecule_type=molecule_type)
        if isinstance(alternate_sequence, str)
        else alternate_sequence
    )

    alignment = generate_alignment(ref_seq_record, alt_seq_record, molecule_type=molecule_type, aligner=aligner)
    variant_blocks = generate_variant_blocks(
        alignment,
        split_consecutive_mismatches=use_non_standard_substitution_rules,
    )
    builder = BUILDER_CONFIG[molecule_type](alignment)
    return [
        builder.build(
            variant_block,
            categorize_variant_block(variant_block, alignment),
        )
        for variant_block in variant_blocks
    ]



from palamedes.hgvs.utils import categorize_variant_block
from palamedes.config import (
    HGVS_VARIANT_TYPE_SUBSTITUTION,
    HGVS_VARIANT_TYPE_DELETION,
    HGVS_VARIANT_TYPE_EXTENSION,
    HGVS_VARIANT_TYPE_DUPLICATION,
    HGVS_VARIANT_TYPE_REPEAT,
    HGVS_VARIANT_TYPE_INSERTION,
    HGVS_VARIANT_TYPE_DELETION_INSERTION,
    VARIANT_BASE_MATCH,
    VARIANT_BASE_MISMATCH,
    VARIANT_BASE_DELETION,
    VARIANT_BASE_INSERTION,
)
from palamedes.models import Block, VariantBlock
from tests.base import PalamedesBaseCase


class CategorizeVariantBlockTestCase(PalamedesBaseCase):
    def test_categorize_variant_block_match_error(self):
        variant_block = VariantBlock(
            Block(0, 1, VARIANT_BASE_MATCH),
            [Block(0, 1, "A")],
            [Block(0, 1, "A")],
        )

        alignment = self.make_alignment("A", "A")

        with self.assertRaisesRegex(ValueError, "with matching bases"):
            categorize_variant_block(variant_block, alignment)

    def test_categorize_variant_block_single_substitution(self):
        variant_block = VariantBlock(
            Block(0, 1, VARIANT_BASE_MISMATCH),
            [Block(0, 1, "A")],
            [Block(0, 1, "T")],
        )

        alignment = self.make_alignment("A", "T")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_SUBSTITUTION)

    def test_categorize_variant_block_single_base_deletion(self):
        variant_block = VariantBlock(
            Block(0, 1, VARIANT_BASE_DELETION),
            [Block(0, 1, "A")],
            [],
        )

        alignment = self.make_alignment("ATTT", "-TTT")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_DELETION)

    def test_categorize_variant_block_multi_base_deletion(self):
        variant_block = VariantBlock(
            Block(0, 3, VARIANT_BASE_DELETION * 3),
            [Block(0, 3, "AAA")],
            [],
        )

        alignment = self.make_alignment("AAACCC", "---CCC")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_DELETION)

    def test_categorize_variant_block_upstream_insertion(self):
        variant_block = VariantBlock(
            Block(0, 3, VARIANT_BASE_INSERTION * 3),
            [],
            [Block(0, 3, "AAA")],
        )

        alignment = self.make_alignment("---CCC", "AAACCC")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_EXTENSION)

    def test_categorize_variant_block_downstream_insertion(self):
        """
        Assume the following fake alignment:

            T T T T - - -
            - - - - A A A
           0 1 2 3 4 5 6 7 ALIGNMENT
           0 1 2 3 4 . . . REF
           . . . . 0 1 2 3 ALT
        """
        variant_block = VariantBlock(
            Block(4, 7, VARIANT_BASE_INSERTION * 3),
            [],
            [Block(0, 3, "AAA")],
        )

        alignment = self.make_alignment("TTTT---", "----AAA")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_EXTENSION)

    def test_categorize_variant_block_downstream_insertion_with_other_gaps(self):
        """
        Assume the following fake alignment:

            T - T T - - -
            T G T T A A A
           0 1 2 3 4 5 6 7 ALIGNMENT
           0 1 . 2 3 . . . REF
           0 1 2 3 4 5 6 7 ALT
        """
        variant_block = VariantBlock(
            Block(4, 7, VARIANT_BASE_INSERTION * 3),
            [],
            [Block(4, 7, "AAA")],
        )

        alignment = self.make_alignment("T-TT---", "TGTTAAA")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_EXTENSION)

    def test_categorize_variant_block_duplication(self):
        """
        Using the following alignment:
            T T T T - - - - A T G
           0 1 2 3 4 5 6 7 8 9 A B
        Set the insert to TTTT, which is directly upstream so duplication
        """
        variant_block = VariantBlock(
            Block(4, 8, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(0, 4, "TTTT")],
        )

        alignment = self.make_alignment("TTTT----ATG", "TTTTTTTTATG")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_DUPLICATION)

    def test_categorize_variant_block_repeat(self):
        """
        Using the following alignment:
            T T T A - - - - A T G
           0 1 2 3 4 5 6 7 8 9 A B
        Set the insert to TATA, TA  is directly upstream so repeat
        """
        variant_block = VariantBlock(
            Block(4, 8, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(0, 4, "TATA")],
        )

        alignment = self.make_alignment("TTTA----ATG", "TTTATATAATG")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_REPEAT)

    def test_categorize_variant_block_insert(self):
        """
        Using the following alignment:
            T T T A - - - - A T G
           0 1 2 3 4 5 6 7 8 9 A B
        Set the insert to GCTA, just an insertion
        """
        variant_block = VariantBlock(
            Block(4, 8, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(0, 4, "GCTA")],
        )

        alignment = self.make_alignment("TTTA----ATG", "TTTAGCTAATG")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_INSERTION)

    def test_categorize_variant_block_double_mismatch(self):
        """2 mismatches is a delins"""
        variant_block = VariantBlock(
            Block(0, 2, VARIANT_BASE_MISMATCH * 2),
            [Block(0, 2, "AA")],
            [Block(0, 2, "CG")],
        )

        alignment = self.make_alignment("AAT", "CGT")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_DELETION_INSERTION)

    def test_categorize_variant_block_mixture(self):
        """
        Using the following alignment:
            A - A T
            - G A T
           0 1 2 3 4
        """
        variant_block = VariantBlock(
            Block(0, 2, VARIANT_BASE_DELETION + VARIANT_BASE_INSERTION),
            [Block(0, 1, "A")],
            [Block(0, 1, "G")],
        )

        alignment = self.make_alignment("A-AT", "-GAT")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_DELETION_INSERTION)

    def test_categorize_variant_block_disrupted_duplication(self):
        """
        Test that upstream gaps are not considered when identifying a duplication, in other words the
        unaligned reference sequence should be used for checking the match. Using the following alignment:

            T - T T T - - - - A T G
            T G T T T T T T T A T G
           0 1 2 3 4 5 6 7 8 9 A B C ALIGN
           0 1 . 2 3 4 . . . . 5 6 7 REF
           0 1 2 3 4 5 6 7 8 9 A B C ALT

        This should be called as a duplication, the upstream reference sequence in the alignment is: -TTT,
        but in the actual sequence it is: TTTT, and that is what should be checked, NOT the version with gaps.
        """
        variant_block = VariantBlock(
            Block(5, 9, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(5, 9, "TTTT")],
        )

        alignment = self.make_alignment("T-TTT----ATG", "TGTTTTTTTATG")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_DUPLICATION)

    def test_categorize_variant_block_disrupted_repeat(self):
        """
        Test that upstream gaps are not considered when identifying a repeat, in other words the
        unaligned reference sequence should be used for checking the match. Using the following alignment:

            T T - A - - - - A T G
            T T G A T A T A A T G
           0 1 2 3 4 5 6 7 8 9 A B ALIGN
           0 1 . 2 3 . . . . 4 5 6 REF
           0 1 2 3 4 5 6 7 8 9 A B ALT

        This should be called as a repeat, the upstream reference sequence in the alignment is: TT-A,
        but in the actual sequence it is: TTA, and that is what should be checked, NOT the version with gaps.
        The insert is TATA, which is a 2 repeat of the TA.
        """
        variant_block = VariantBlock(
            Block(5, 9, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(5, 9, "TATA")],
        )

        alignment = self.make_alignment("TT-A----ATG", "TTGATATAATG")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_REPEAT)

    def test_categorize_variant_block_out_of_bounds_upstream(self):
        """
        Ensure the upstream sequence check is safe in terms of looking out of bounds, using the following
        alignment:
            A - - - - A T G
            A T T T T A T G
           0 1 2 3 4 5 6 7 8 ALIGN
           0 1 . . . . 2 3 4 REF
           0 1 2 3 4 5 6 7 8 ALT
        """
        variant_block = VariantBlock(
            Block(1, 5, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(1, 5, "TTTT")],
        )

        alignment = self.make_alignment("A----ATG", "ATTTTATG")

        self.assertEqual(categorize_variant_block(variant_block, alignment), HGVS_VARIANT_TYPE_INSERTION)



from hgvs.sequencevariant import SequenceVariant

from palamedes.hgvs.builders import HgvsProteinBuilder
from palamedes.models import VariantBlock, Block
from palamedes.config import (
    VARIANT_BASE_MISMATCH,
    HGVS_VARIANT_TYPE_SUBSTITUTION,
    REF_SEQUENCE_ID,
    HGVS_TYPE_PROTEIN,
    HGVS_VARIANT_TYPE_DELETION,
    VARIANT_BASE_DELETION,
    VARIANT_BASE_INSERTION,
    HGVS_VARIANT_TYPE_INSERTION,
    HGVS_VARIANT_TYPE_EXTENSION,
    HGVS_VARIANT_TYPE_REPEAT,
    HGVS_VARIANT_TYPE_DUPLICATION,
    HGVS_VARIANT_TYPE_DELETION_INSERTION,
)

from tests.base import PalamedesBaseCase


class HgvsProteinBuilderTestCase(PalamedesBaseCase):
    def assert_variant_string_matches(self, hgvs_variant: SequenceVariant, variant_string: str) -> None:
        self.assertEqual(
            hgvs_variant.format(conf={"p_3_letter": False}),
            f"{REF_SEQUENCE_ID}:{HGVS_TYPE_PROTEIN}.{variant_string}",
        )

    def test_hgvs_protein_builder_build_substitution(self):
        alignment = self.make_alignment("FFF", "FSF")
        variant_block = VariantBlock(
            Block(1, 2, VARIANT_BASE_MISMATCH),
            [Block(1, 2, "F")],
            [Block(1, 2, "S")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_SUBSTITUTION)
        self.assert_variant_string_matches(as_hgvs, "F2S")

    def test_hgvs_protein_builder_build_substitution_first_position(self):
        alignment = self.make_alignment("FFF", "SFF")
        variant_block = VariantBlock(
            Block(0, 1, VARIANT_BASE_MISMATCH),
            [Block(0, 1, "F")],
            [Block(0, 1, "S")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_SUBSTITUTION)
        self.assert_variant_string_matches(as_hgvs, "F1S")

    def test_hgvs_protein_builder_build_substitution_last_position(self):
        alignment = self.make_alignment("FFF", "FFS")
        variant_block = VariantBlock(
            Block(2, 3, VARIANT_BASE_MISMATCH),
            [Block(2, 3, "F")],
            [Block(2, 3, "S")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_SUBSTITUTION)
        self.assert_variant_string_matches(as_hgvs, "F3S")

    def test_hgvs_protein_builder_build_deletion_single_base(self):
        alignment = self.make_alignment("FFF", "F-F")
        variant_block = VariantBlock(
            Block(1, 2, VARIANT_BASE_DELETION),
            [Block(1, 2, "F")],
            [],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DELETION)
        self.assert_variant_string_matches(as_hgvs, "F2del")

    def test_hgvs_protein_builder_build_deletion_multi_base(self):
        """
         F S S S S F
         F - - - - F
        0 1 2 3 4 5 6
        """
        alignment = self.make_alignment("FSSSSF", "F----F")
        variant_block = VariantBlock(
            Block(1, 5, VARIANT_BASE_DELETION * 4),
            [Block(1, 5, "S" * 4)],
            [],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DELETION)
        self.assert_variant_string_matches(as_hgvs, "S2_S5del")

    def test_hgvs_protein_builder_build_deletion_multi_base_covering_start(self):
        """
         S S S S F
         - - - - F
        0 1 2 3 4 5
        """
        alignment = self.make_alignment("SSSSF", "----F")
        variant_block = VariantBlock(
            Block(0, 4, VARIANT_BASE_DELETION * 4),
            [Block(0, 4, "S" * 4)],
            [],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DELETION)
        self.assert_variant_string_matches(as_hgvs, "S1_S4del")

    def test_hgvs_protein_builder_build_deletion_multi_base_covering_end(self):
        """
         F S S S S
         F - - - -
        0 1 2 3 4 5
        """
        alignment = self.make_alignment("FSSSS", "F----")
        variant_block = VariantBlock(
            Block(1, 5, VARIANT_BASE_DELETION * 4),
            [Block(1, 5, "S" * 4)],
            [],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DELETION)
        self.assert_variant_string_matches(as_hgvs, "S2_S5del")

    def test_hgvs_protein_builder_build_insertion_single_base(self):
        alignment = self.make_alignment("F-F", "FSF")
        variant_block = VariantBlock(
            Block(1, 2, VARIANT_BASE_INSERTION),
            [],
            [Block(1, 2, "S")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_INSERTION)
        self.assert_variant_string_matches(as_hgvs, "F1_F2insS")

    def test_hgvs_protein_builder_build_insertion_multiple_base(self):
        """
         F - - - - - F
         F S S L Q N F
        0 1 2 3 4 5 6 7
        """
        alignment = self.make_alignment("F-----F", "FSSLQNF")
        variant_block = VariantBlock(
            Block(1, 6, VARIANT_BASE_INSERTION * 5),
            [],
            [Block(1, 6, "SSLQNF")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_INSERTION)
        self.assertEqual(
            as_hgvs.format(conf={"p_3_letter": False}), f"{REF_SEQUENCE_ID}:{HGVS_TYPE_PROTEIN}.F1_F2insSSLQNF"
        )

    def test_hgvs_protein_builder_build_extension_upstream_single_base(self):
        """
         - F F
         S F F
        0 1 2 3
        """
        alignment = self.make_alignment("-FF", "SFF")
        variant_block = VariantBlock(
            Block(0, 1, VARIANT_BASE_INSERTION),
            [],
            [Block(0, 1, "S")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_EXTENSION)
        self.assert_variant_string_matches(as_hgvs, "F1extS-1")

    def test_hgvs_protein_builder_build_extension_upstream_multiple_base(self):
        """
         - - - F F
         S K L F F
        0 1 2 3 5 6
        """
        alignment = self.make_alignment("---FF", "SKLFF")
        variant_block = VariantBlock(
            Block(0, 3, VARIANT_BASE_INSERTION * 3),
            [],
            [Block(0, 3, "SKL")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_EXTENSION)
        self.assertEqual(
            as_hgvs.format(conf={"p_3_letter": False}), f"{REF_SEQUENCE_ID}:{HGVS_TYPE_PROTEIN}.F1extSKL-3"
        )

    def test_hgvs_protein_builder_build_extension_downstream_single_base(self):
        """
         F F -
         F F S
        0 1 2 3
        """
        alignment = self.make_alignment("FF-", "FFS")
        variant_block = VariantBlock(
            Block(2, 3, VARIANT_BASE_INSERTION),
            [],
            [Block(2, 3, "S")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_EXTENSION)
        self.assert_variant_string_matches(as_hgvs, "F2extS1")

    def test_hgvs_protein_builder_build_extension_downstream_multiple_bases(self):
        """
         F F - - - -
         F F S L L Q
        0 1 2 3 4 5 6
        """
        alignment = self.make_alignment("FF----", "FFSLLQ")
        variant_block = VariantBlock(
            Block(2, 6, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(2, 6, "SLLQ")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_EXTENSION)
        self.assertEqual(
            as_hgvs.format(conf={"p_3_letter": False}), f"{REF_SEQUENCE_ID}:{HGVS_TYPE_PROTEIN}.F2extSLLQ4"
        )

    def test_hgvs_protein_builder_build_duplication_single_base(self):
        """
         F F - A
         F F F A
        0 1 2 3 4
        """
        alignment = self.make_alignment("FF-A", "FFFA")
        variant_block = VariantBlock(
            Block(2, 3, VARIANT_BASE_INSERTION),
            [],
            [Block(2, 3, "F")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DUPLICATION)
        self.assert_variant_string_matches(as_hgvs, "F2dup")

    def test_hgvs_protein_builder_build_duplication_multiple_bases(self):
        """
         F F S A - - - - K
         F F S A F F S A K
        0 1 2 3 4 5 6 7 8 9
        """
        alignment = self.make_alignment("FFSA----K", "FFSAFFSAK")
        variant_block = VariantBlock(
            Block(4, 8, VARIANT_BASE_INSERTION * 4),
            [],
            [Block(4, 8, "FFSA")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DUPLICATION)
        self.assert_variant_string_matches(as_hgvs, "F1_A4dup")

    def test_hgvs_protein_builder_build_repeat_single_base(self):
        """
         F - - A
         F F F A
        0 1 2 3 4
        """
        alignment = self.make_alignment("F--A", "FFFA")
        variant_block = VariantBlock(
            Block(1, 3, VARIANT_BASE_INSERTION * 2),
            [],
            [Block(1, 3, "FF")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_REPEAT)
        self.assert_variant_string_matches(as_hgvs, "F1[2]")

    def test_hgvs_protein_builder_build_repeat_multiple_bases(self):
        """
         F L S - - - - - - - - - A
         F L S F L S F L S F L S A
        0 1 2 3 4 5 6 7 8 9 A B C D
        """
        alignment = self.make_alignment("FLS---------A", "FLSFLSFLSFLSA")
        variant_block = VariantBlock(
            Block(3, 12, VARIANT_BASE_INSERTION * 9),
            [],
            [Block(3, 12, "FLSFLSFLS")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_REPEAT)
        self.assert_variant_string_matches(as_hgvs, "F1_S3[3]")

    def test_hgvs_protein_builder_build_deletion_insertion(self):
        """
         F L S -
         F K - A
        0 1 2 3 4 ALIGNMENT
        0 1 2 3 . REF
        0 1 . 2 3 ALT
        """
        alignment = self.make_alignment("FLS-", "FK-A")
        variant_block = VariantBlock(
            Block(1, 4, VARIANT_BASE_MISMATCH + VARIANT_BASE_DELETION + VARIANT_BASE_INSERTION),
            [Block(1, 3, "LS")],
            [Block(1, 3, "KA")],
        )

        as_hgvs = HgvsProteinBuilder(alignment).build(variant_block, HGVS_VARIANT_TYPE_DELETION_INSERTION)
        self.assert_variant_string_matches(as_hgvs, "L2_S3delinsKA")



from setuptools import find_packages, setup

version = "0.0.7"

setup(
    name="palamedes",
    version=version,
    url="https://github.com/mammothbio-os/palamedes",
    python_requires=">=3.11",
    description="Palamedes: HGVS variants from a sequence alignment",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author_email="open-source@mammothbiosci.com",
    entry_points={
        "console_scripts": ["palamedes=palamedes.__main__:main"],
    },
    packages=find_packages(),
    install_requires=["biopython~=1.83", "hgvs~=1.5.4"],
    keywords=[
        "bioinformatics",
        "alignment",
        "hgvs",
        "variant",
        "amino acid",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.11",
  ],
)



from typing import NamedTuple


class Block(NamedTuple):
    """
    General purpose block for slices of the alignment. Note (start, end) should be ZBHO
    """

    start: int
    end: int
    bases: str

    @classmethod
    def collapse(cls, blocks: list["Block"]) -> "Block":
        """
        Given a variable length list of blocks, collapse all blocks into one or raise a ValueError if the list is empty.
        Blocks will be sorted but must be adjacent to each-other.
        """
        if len(blocks) == 0:
            raise ValueError("Cannot collapse empty list of Blocks")

        sorted_blocks = sorted(blocks)
        for idx, block in enumerate(sorted_blocks[:-1]):
            if sorted_blocks[idx + 1].start != block.end:
                raise ValueError(f"Cannot collapse Blocks, they must be adjacent. Blocks {idx} and {idx + 1} are not!")

        new_start = sorted_blocks[0].start
        new_end = sorted_blocks[-1].end
        new_bases = "".join([block.bases for block in sorted_blocks])
        return Block(new_start, new_end, new_bases)


class VariantBlock(NamedTuple):
    """
    Internal representation of a slice of an alignment, based on a contiguous
    run of positions that are not matches. The object tracks the start and
    end positions from each sequence which are covered by the block, as well
    as the bases. It also tracks the start and end of the block within the
    alignment and the "variant bases", which is similar to a CIGAR string
    but without the short-hand notation.

    Note that:
    - reference_blocks and alternate_blocks store Block objects corresponding to the block from either
      sequence in the alignment. The list should have 0 or 1 elements. The 0 is if no part of the sequence
      is included in the alignment (like an insertion upstream of the first base).
    - Alignment block will always be set, and should have a sequence with the VARIANT_BASE characters
    """

    alignment_block: Block
    reference_blocks: list[Block]
    alternate_blocks: list[Block]



from unittest import TestCase

from palamedes.models import Block


class BlockTestCase(TestCase):
    def test_block_create(self):
        bases = "AAAA"
        start = 0
        end = 4
        block = Block(start, end, bases)

        self.assertEqual(block.bases, bases)
        self.assertEqual(block.start, start)
        self.assertEqual(block.end, end)

    def test_block_collapse_empty_error(self):
        with self.assertRaisesRegex(ValueError, "empty list"):
            Block.collapse([])

    def test_block_collapse_not_adjacent_error(self):
        first_block = Block(0, 4, "A" * 4)
        second_block = Block(5, 10, "G" * 6)

        with self.assertRaisesRegex(ValueError, "they must be adjacent."):
            Block.collapse([first_block, second_block])

    def test_block_collapse_single(self):
        block = Block(0, 4, "A" * 4)
        collapsed_block = Block.collapse([block])
        self.assertEqual(collapsed_block, block)

    def test_block_collapse(self):
        first_block = Block(0, 4, "A" * 4)
        second_block = Block(4, 10, "G" * 6)
        third_block = Block(10, 15, "C" * 5)

        collapsed_block = Block.collapse([first_block, second_block, third_block])
        self.assertEqual(collapsed_block.start, first_block.start)
        self.assertEqual(collapsed_block.end, third_block.end)
        self.assertEqual(collapsed_block.bases, "A" * 4 + "G" * 6 + "C" * 5)

    def test_block_collapse_unsorted(self):
        first_block = Block(0, 4, "A" * 4)
        second_block = Block(4, 10, "G" * 6)
        third_block = Block(10, 15, "C" * 5)

        # shuffle the order
        collapsed_block = Block.collapse([third_block, first_block, second_block])
        self.assertEqual(collapsed_block.start, first_block.start)
        self.assertEqual(collapsed_block.end, third_block.end)
        self.assertEqual(collapsed_block.bases, "A" * 4 + "G" * 6 + "C" * 5)
